#include <time.h>
#include "dcopynt.h"
#include "bootmsg.h"

#define TRACKSIZE(dg) (dg.SectorsPerTrack*dg.BytesPerSector)
#define BUFPOS(x,y,dg) ((x*dg.TracksPerCylinder+y)*TRACKSIZE(dg))
#define BUFSIZE(dg) (dg.BytesPerSector*(long)dg.Cylinders.QuadPart*dg.TracksPerCylinder*dg.SectorsPerTrack)
#define FILEPOS(x,y,s,dg) (BUFPOS(x,y,dg)+s*dg.BytesPerSector)

char *buffer=NULL;
int *fulltrack=NULL;
int startsec;
DISK_GEOMETRY g_dg;
float g_progress;
int sector_spin=SECTOR_SPIN;

void DisplayStatus(HWND hWnd,char *str)
{
  SendDlgItemMessage(hWnd,IDC_STATUS,WM_SETTEXT,0,(LPARAM)str);
  SetWindowText(hWnd,str);
}

void DisplayProgress(HWND hWnd,int track,int head,int attempt,DISK_GEOMETRY dg)
{
  float progress,total,current;
  char str[10];

  total=(float)dg.Cylinders.QuadPart*dg.TracksPerCylinder;
  current=track*(float)dg.TracksPerCylinder+head;
  sprintf(str,"%d",track);
  SendDlgItemMessage(hWnd,IDC_TRACK,WM_SETTEXT,0,(LPARAM)str);
  sprintf(str,"%d",head);
  SendDlgItemMessage(hWnd,IDC_SIDE,WM_SETTEXT,0,(LPARAM)str);
  sprintf(str,"%d",attempt+1);
  SendDlgItemMessage(hWnd,IDC_ATTEMPT,WM_SETTEXT,0,(LPARAM)str);
  progress=current*100/total;
  SendDlgItemMessage(hWnd,IDC_PROGRESS,PBM_SETPOS,(WPARAM)progress,0);
  g_progress=progress;
}

void DisplayETA(HWND hWnd)
{
  float elapsed,remaining;
  char str[10];

  if (startsec!=0)
  {
    elapsed=(float)(time(NULL)-startsec);
    if (g_progress)
    {
      remaining=(100-g_progress)/g_progress*elapsed;
      sprintf(str,"%02d:%02d",(int)remaining/60,(int)remaining%60);
      SendDlgItemMessage(hWnd,IDC_ETA,WM_SETTEXT,0,(LPARAM)str);
    }
  }
}

void Busy(HWND hWnd,BOOL on)
{
  HCURSOR hc;
  static HCURSOR oldhc;

  if (on)
  {
    hc=LoadCursor(NULL,IDC_WAIT);
    SetCapture(hWnd);
    oldhc=SetCursor(hc);
  }
  else
  {
    SetCursor(oldhc);
    ReleaseCapture();
  }
}

void DisplayLastError(HWND hWnd)
{
  LPVOID lpMsgBuf;
  
  FormatMessage( 
    FORMAT_MESSAGE_ALLOCATE_BUFFER | FORMAT_MESSAGE_FROM_SYSTEM,
    NULL,
    GetLastError(),
    MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT),
    (LPTSTR) &lpMsgBuf,    0,    NULL );
  Error( hWnd, lpMsgBuf);
  LocalFree( lpMsgBuf );
}

int AskLastError(HWND hWnd)
{
  LPVOID lpMsgBuf;
  int ret;
  
  FormatMessage( 
    FORMAT_MESSAGE_ALLOCATE_BUFFER | FORMAT_MESSAGE_FROM_SYSTEM,
    NULL,
    GetLastError(),
    MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT),
    (LPTSTR) &lpMsgBuf,    0,    NULL );
  ret=MessageBox(hWnd,lpMsgBuf,"Disk Error",MB_ABORTRETRYIGNORE|MB_ICONSTOP|MB_DEFBUTTON2);
  LocalFree( lpMsgBuf );
  return ret;
}

BOOL AllocateMemory(HWND hWnd,DISK_GEOMETRY dg)
{
  if (buffer)
    GlobalFree(buffer);
  if ((buffer=GlobalAlloc(GMEM_FIXED,BUFSIZE(dg)))==NULL)
  {
    DisplayLastError(hWnd);
    return FALSE;
  }
  EnableWindow(GetDlgItem(g_hWnd,IDC_WRITE),TRUE);
  EnableWindow(GetDlgItem(g_hWnd,IDC_VERIFY),TRUE);
  return TRUE;
}

void FreeMemory(void)
{
  if (buffer)
    GlobalFree(buffer);
  EnableWindow(GetDlgItem(g_hWnd,IDC_WRITE),FALSE);
  EnableWindow(GetDlgItem(g_hWnd,IDC_VERIFY),FALSE);
  buffer=NULL;
}

HANDLE OpenInputFile(HWND hWnd,char *filename)
{
  HANDLE h;

  h=CreateFile(filename,GENERIC_READ,FILE_SHARE_READ,NULL,OPEN_EXISTING,FILE_ATTRIBUTE_NORMAL,NULL);
  if (h==INVALID_HANDLE_VALUE)
  {
    DisplayLastError(hWnd);
    return NULL;
  }
  return h;
}

HANDLE OpenOutputFile(HWND hWnd,char *filename,int mode)
{
  HANDLE h;

  if (verify_data)
    h=CreateFile(filename,GENERIC_WRITE|GENERIC_READ,FILE_SHARE_READ|FILE_SHARE_WRITE,NULL,mode,FILE_ATTRIBUTE_NORMAL,NULL);
  else
    h=CreateFile(filename,GENERIC_WRITE,FILE_SHARE_WRITE,NULL,mode,FILE_ATTRIBUTE_NORMAL,NULL);
  if (h==INVALID_HANDLE_VALUE)
  {
    DisplayLastError(hWnd);
    return NULL;
  }
  return h;
}

void DisplayDiskInfo(HWND hWnd,DISK_GEOMETRY dg)
{
  char str[100];

  sprintf(str,"%d",(int)dg.Cylinders.QuadPart);
  SendDlgItemMessage(hWnd,IDC_TRACKS,WM_SETTEXT,0,(LPARAM)str);
  sprintf(str,"%d",(int)dg.TracksPerCylinder);
  SendDlgItemMessage(hWnd,IDC_SIDES,WM_SETTEXT,0,(LPARAM)str);
  sprintf(str,"%d",(int)dg.SectorsPerTrack);
  SendDlgItemMessage(hWnd,IDC_SECTORS,WM_SETTEXT,0,(LPARAM)str);
  sprintf(str,"%d",(int)dg.BytesPerSector);
  SendDlgItemMessage(hWnd,IDC_BYTES,WM_SETTEXT,0,(LPARAM)str);
  sprintf(str,"%ld K",(long)(dg.BytesPerSector*dg.Cylinders.QuadPart*dg.TracksPerCylinder*dg.SectorsPerTrack/1024L));
  SendDlgItemMessage(hWnd,IDC_TOTALSIZE,WM_SETTEXT,0,(LPARAM)str);
}

int GetDiskInfo(HWND hWnd,HANDLE h,DISK_GEOMETRY *dg)
{
  DWORD sz;

  if (!DeviceIoControl(
    h,
    IOCTL_DISK_GET_DRIVE_GEOMETRY,
    NULL,
    0,
    dg,
    sizeof(DISK_GEOMETRY),
    &sz,
    NULL))
  {
    DisplayLastError(hWnd);
    return 0;
  }
  dg->MediaType=0;
  return 1;
}

int GuessDiskImageFormat(HANDLE h,DISK_GEOMETRY *dg)
{
  int tracks,sectors,sides;

  switch (GetFileSize(h,NULL))
  {
    case 163840L:
      tracks=40;
      sectors=8;
      sides=1;
      break;
    case 184320L:
      tracks=40;
      sectors=9;
      sides=1;
      break;
    case 327680L:
      tracks=40;
      sectors=8;
      sides=2;
      break;
    case 368640L:
      tracks=40;
      sectors=9;
      sides=2;
      break;
    case 737280L:
      tracks=80;
      sectors=9;
      sides=2;
      break;
    case 819200L:
      tracks=80;
      sectors=10;
      sides=2;
      break;
    case 1228800L:
      tracks=80;
      sectors=15;
      sides=2;
      break;
    case 1474560L:
      tracks=80;
      sectors=18;
      sides=2;
      break;
    case 1720320L:
      tracks=80;
      sectors=21;
      sides=2;
      break;
    default:
      return 0;
  }
  dg->Cylinders.QuadPart=tracks;
  dg->TracksPerCylinder=sides;
  dg->SectorsPerTrack=sectors;
  dg->BytesPerSector=512;
  dg->MediaType=0;
  return 1;
}

int GetDiskImageFormat(HWND hWnd,HANDLE h,DISK_GEOMETRY *dg)
{
  char fatinfo[512];
  WORD total_sectors;
  DWORD bytes_read;

  if (!ReadFile(h,fatinfo,512,&bytes_read,NULL))
    return 0;
  dg->MediaType=0;
  dg->BytesPerSector=*(WORD*)&fatinfo[11];
  total_sectors=*(WORD*)&fatinfo[19];
  dg->SectorsPerTrack=*(WORD*)&fatinfo[24];
  dg->TracksPerCylinder=*(WORD*)&fatinfo[26];
  if (dg->SectorsPerTrack==0 || dg->TracksPerCylinder==0)
  {
    if (!GuessDiskImageFormat(h,&w_dg))
    {
      if (!AskDiskFormat(hWnd,0))
        return 0;
    }
    *dg=w_dg;
    g_dg=w_dg;
    SetFilePointer(h,0,NULL,FILE_BEGIN);
    return 1;
  }
  dg->Cylinders.QuadPart=total_sectors/dg->SectorsPerTrack/dg->TracksPerCylinder;
  if (dg->Cylinders.QuadPart==0)
  {
    if (!GuessDiskImageFormat(h,&w_dg))
    {
      if (!AskDiskFormat(hWnd,0))
        return 0;
    }
    *dg=w_dg;
    g_dg=w_dg;
    SetFilePointer(h,0,NULL,FILE_BEGIN);
    return 1;
  }
  g_dg=*dg;
  SetFilePointer(h,0,NULL,FILE_BEGIN);
  return 1;
}

int ReadTrack(HANDLE h,int track,int side,char *buf,DISK_GEOMETRY dg)
{
  static int sector=0;
  DWORD bytes_read;

  sector=(sector+sector_spin)%dg.SectorsPerTrack;
  if (sector)
  {
    if (SetFilePointer(h,FILEPOS(track,side,sector,dg),NULL,FILE_BEGIN)==0xFFFFFFFF)
      return 0;
    if (!ReadFile(h,&buf[sector*dg.BytesPerSector],(dg.SectorsPerTrack-sector)*dg.BytesPerSector,&bytes_read,NULL))
      return 0;
    if (SetFilePointer(h,BUFPOS(track,side,dg),NULL,FILE_BEGIN)==0xFFFFFFFF)
      return 0;
    if (!ReadFile(h,buf,sector*dg.BytesPerSector,&bytes_read,NULL))
      return 0;
  }
  else
  {
    if (SetFilePointer(h,BUFPOS(track,side,dg),NULL,FILE_BEGIN)==0xFFFFFFFF)
      return 0;
    if (!ReadFile(h,buf,dg.SectorsPerTrack*dg.BytesPerSector,&bytes_read,NULL))
      return 0;
  }
  return 1;
}
  
int WriteTrack(HANDLE h,int track,int side,char *buf)
{
  static int sector=0;
  DWORD bytes_read;

  sector=(sector+sector_spin)%g_dg.SectorsPerTrack;
  if (sector)
  {
    if (SetFilePointer(h,FILEPOS(track,side,sector,g_dg),NULL,FILE_BEGIN)==0xFFFFFFFF)
      return 0;
    if (!WriteFile(h,&buf[sector*g_dg.BytesPerSector],(g_dg.SectorsPerTrack-sector)*g_dg.BytesPerSector,&bytes_read,NULL))
      return 0;
    if (SetFilePointer(h,BUFPOS(track,side,g_dg),NULL,FILE_BEGIN)==0xFFFFFFFF)
      return 0;
    if (!WriteFile(h,buf,sector*g_dg.BytesPerSector,&bytes_read,NULL))
      return 0;
  }
  else
  {
    if (SetFilePointer(h,BUFPOS(track,side,g_dg),NULL,FILE_BEGIN)==0xFFFFFFFF)
      return 0;
    if (!WriteFile(h,buf,g_dg.SectorsPerTrack*g_dg.BytesPerSector,&bytes_read,NULL))
      return 0;
  }
  return 1;
}

void GetEmptyTracks(unsigned char *buf,DISK_GEOMETRY dg)
{
  int x,track_size,start,value,sectors_per_cluster,sect_in_res_area;
  int fat_copies,root_entries,total_sectors,sect_per_fat,sect_per_track;
  int heads,hidden_sectors,tracks;

  if (fulltrack)
    GlobalFree(fulltrack);
  fulltrack=(int*)GlobalAlloc(GMEM_FIXED,(int)dg.Cylinders.QuadPart*sizeof(int));
  memset(fulltrack,1,(int)dg.Cylinders.QuadPart*sizeof(int));
  sectors_per_cluster=(int)*(BYTE*)&buf[13];
  sect_in_res_area=*(WORD*)&buf[14];
  fat_copies=*(BYTE*)&buf[16];
  root_entries=*(WORD*)&buf[17];
  total_sectors=*(WORD*)&buf[19];
  sect_per_fat=*(WORD*)&buf[22];
  sect_per_track=*(WORD*)&buf[24];
  heads=*(WORD*)&buf[26];
  hidden_sectors=*(WORD*)&buf[28];
  if (sect_per_track==0 || heads==0)
    return;
  tracks=total_sectors/sect_per_track/heads;
  if (tracks==0)
    return;
  memset(fulltrack,0,(int)dg.Cylinders.QuadPart*sizeof(int));
  start=sect_in_res_area+sect_per_fat*fat_copies+root_entries*32/512;
  track_size=sect_in_res_area*512;
  for (x=0;x<(total_sectors-start)/sectors_per_cluster;x++)
  {
    if ((x*3)&1)
      value=(*(unsigned*)(buf+track_size+3+x*3/2))>>4;
    else
      value=(*(unsigned*)(buf+track_size+3+x*3/2))&0x0FFF;
    if (value && value!=0x0ff7)
      fulltrack[(x*sectors_per_cluster+start)/(sect_per_track*heads)]=1;
  }
  fulltrack[0]=1;
}

void GetPars(DISK_GEOMETRY dg,int *GAP,int *interleave,int *sect_per_fat,int *root_entries)
{
  switch (dg.SectorsPerTrack)
  {
     case 19:case 20:(*GAP)=0x2A;(*interleave)=1;*sect_per_fat=3;*root_entries=224;break;
     case 21:case 22:(*GAP)=0x2;(*interleave)=2;*sect_per_fat=3;*root_entries=16;break;
     case 1:case 2:case 3:case 4:case 5:case 6 :case 7:(*GAP)=0x50;(*interleave)=1;*sect_per_fat=3;*root_entries=16;break;
     case 8:case 9:(*GAP)=0x50;(*interleave)=1;*sect_per_fat=6;*root_entries=56;break;
     case 10:case 11:case 12:(*GAP)=0x02;(*interleave)=2;*sect_per_fat=6;*root_entries=112;break;
     case 13:case 14:case 15:case 16:(*GAP)=0x50;(*interleave)=1;*sect_per_fat=9;*root_entries=112;break;
     default:(*GAP)=0x6C;(*interleave)=1;*sect_per_fat=9;*root_entries=224;break;
  };
}

int WriteLogical(HANDLE h,HWND hWnd)
{
  int  TotalSectors,SectorsPerFAT,GAP,Interleave,RootEntries;
  char *TrakBuffer;
  DWORD bytes_read;

  if (stopthread1)
    return 1;
  DisplayStatus(hWnd,"Writing disk info");
  GetPars(w_dg,&GAP,&Interleave,&SectorsPerFAT,&RootEntries);
  if (NULL==(TrakBuffer=(char*)LocalAlloc(LMEM_FIXED,max(w_dg.SectorsPerTrack,9)*w_dg.BytesPerSector)))
  {
    DisplayLastError(hWnd);
    return 0;
  }
  memset( &TrakBuffer[0],0,w_dg.SectorsPerTrack*w_dg.BytesPerSector );
  memcpy( &TrakBuffer[0],BootMask,102);
  memcpy( &TrakBuffer[3],"DCOPYNT1",8);
  memcpy( &TrakBuffer[102],BootMes,sizeof(BootMes));
  TrakBuffer[510]=0x55;
  TrakBuffer[511]=(char)0xAA;
  TotalSectors=w_dg.SectorsPerTrack*w_dg.TracksPerCylinder*(int)w_dg.Cylinders.QuadPart;
  TrakBuffer[ 11 ] = (char)(w_dg.BytesPerSector%256);
  TrakBuffer[ 12 ] = (char)(w_dg.BytesPerSector/256);
  TrakBuffer[ 13 ] = (char)((w_dg.SectorsPerTrack>18)?4:1);
  TrakBuffer[ 14 ] = (char)(1%256); // Sectors in res. area
  TrakBuffer[ 15 ] = (char)(1/256); // Sectors in res. area
  TrakBuffer[ 16 ] = (char)2; // FAT copies
  TrakBuffer[ 17 ] = (char)(RootEntries); // root entries
  TrakBuffer[ 19 ] = (char)(TotalSectors%256);
  TrakBuffer[ 20 ] = (char)(TotalSectors/256);
  TrakBuffer[ 21 ] = (char)0xf0; // Media descriptor
  TrakBuffer[ 22 ] = (char)(SectorsPerFAT); // Sectors per FAT
  TrakBuffer[ 24 ] = (char)w_dg.SectorsPerTrack;
  TrakBuffer[ 26 ] = (char)w_dg.TracksPerCylinder;
  TrakBuffer[ 28 ] = (char)0; // Hidden Sectors
  TrakBuffer[ w_dg.BytesPerSector ] = (char)0xf0; // Media descriptor
  TrakBuffer[ w_dg.BytesPerSector+1 ] = (char)0xFF;
  TrakBuffer[ w_dg.BytesPerSector+2 ] = (char)0xFF;
  TrakBuffer[ w_dg.BytesPerSector*(SectorsPerFAT + 1)] = (char)0xf0; // Media descriptor
  TrakBuffer[ w_dg.BytesPerSector*(SectorsPerFAT + 1)+ 1 ] =(char) 0xFF;
  TrakBuffer[ w_dg.BytesPerSector*(SectorsPerFAT + 1)+ 2 ] = (char)0xFF;
  SetFilePointer(h,0,NULL,FILE_BEGIN);
  if (!WriteFile(h,TrakBuffer,w_dg.SectorsPerTrack*w_dg.BytesPerSector,&bytes_read,NULL))
  {
    DisplayLastError(hWnd);
    LocalFree(TrakBuffer);
    return 0;
  }
  if (w_dg.TracksPerCylinder>1)
  {
    memset( &TrakBuffer[0],0,w_dg.SectorsPerTrack*w_dg.BytesPerSector );
    if (!WriteFile(h,TrakBuffer,w_dg.SectorsPerTrack*w_dg.BytesPerSector,&bytes_read,NULL))
    {
      DisplayLastError(hWnd);
      LocalFree(TrakBuffer);
      return 0;
    }
  }
  LocalFree(TrakBuffer);
  return 1;
}

int FormatDisk(HWND hWnd,HANDLE *h,BOOLEAN writelogical)
{
  FORMAT_EX_PARAMETERS *fp;
  DWORD sz;
  int x,y,interleave,gap,ss=0,track,head,sect_per_fat,root_entries,r;
  void *verify_buffer;

  startsec=0;
  ss=0;
  if (!quickformat)
  {
    startsec=time(NULL);
    if (verify_data)
    {
      if ((verify_buffer=GlobalAlloc(GMEM_FIXED,w_dg.SectorsPerTrack*w_dg.BytesPerSector))==NULL)
      {
				DisplayLastError(hWnd);
				ss=SECTOR_SPIN;
				return FALSE;
      }
    }
    GetPars(w_dg,&gap,&interleave,&sect_per_fat,&root_entries);
    fp=(FORMAT_EX_PARAMETERS*)LocalAlloc(LMEM_FIXED,sizeof(FORMAT_EX_PARAMETERS)+w_dg.SectorsPerTrack*sizeof(WORD));
    for (track=0;track<w_dg.Cylinders.QuadPart && !stopthread1;track++)
    {
      for (head=0;head<(int)w_dg.TracksPerCylinder && !stopthread1;head++)
      {
        for (x=0;x<(int)w_dg.SectorsPerTrack;x++)
				{
          y=(x+track*ss)%(int)w_dg.SectorsPerTrack;
          fp->SectorNumber[x]=(y%interleave)*(((int)w_dg.SectorsPerTrack+1)/interleave)+1+y/interleave;
				}
        fp->MediaType=F3_1Pt44_512;
        fp->StartCylinderNumber=track;
        fp->EndCylinderNumber=track;
        fp->StartHeadNumber=head;
        fp->EndHeadNumber=head;
        fp->FormatGapLength=gap;
        fp->SectorsPerTrack=(int)w_dg.SectorsPerTrack;
	      r=0;
        DisplayProgress(hWnd,track,head,r,w_dg);
				while (r>=0 && !stopthread1)
				{
          DisplayStatus(hWnd,"Formatting disk");
					if (!stopthread1 && !DeviceIoControl(
						*h,
						IOCTL_DISK_FORMAT_TRACKS_EX,
						fp,
						sizeof(FORMAT_EX_PARAMETERS)+w_dg.SectorsPerTrack*sizeof(WORD),
						NULL,
						0,
						&sz,
						NULL))
					{
						if (++r>=retries)
						{
							switch (AskLastError(hWnd))
							{
								case IDABORT:
									GlobalFree(fp);
									ss=SECTOR_SPIN;
									return FALSE;
								case IDIGNORE:
									r=-1;
									break;
								case IDRETRY:
									CloseHandle(*h);
									*h=OpenOutputFile(hWnd,"\\\\.\\A:",OPEN_EXISTING);
									if (!(*h))
									{
										if (verify_data)
											GlobalFree(verify_buffer);
										GlobalFree(fp);
																ss=SECTOR_SPIN;
										return FALSE;
									}
									r=0;
									break;
							}
						}
						DisplayProgress(hWnd,track,head,r,w_dg);
						continue;
					}
					else
					{
						if (!verify_data)
							r=-1;
					}
					if (verify_data)
					{
						DisplayStatus(hWnd,"Verifying disk");
						if (!stopthread1 && !ReadTrack(*h,track,head,verify_buffer,w_dg))
						{
							if (++r>=retries)
							{
								switch (AskLastError(hWnd))
								{
									case IDABORT:
										if (verify_data)
											GlobalFree(verify_buffer);
  										ss=SECTOR_SPIN;
										return FALSE;
									case IDIGNORE:
										r=-1;
										break;
									case IDRETRY:
										r=0;
										break;
								}
							}
  						DisplayProgress(hWnd,track,head,r,w_dg);
							continue;
						}
						else
							r=-1;
					}
				}
      }
    }
    GlobalFree(fp);
  }
  if (writelogical)
    WriteLogical(*h,hWnd);
  if (verify_data)
    GlobalFree(verify_buffer);
  ss=SECTOR_SPIN;
  return 1;
}
  
int ReadDisk(HWND hWnd,int ask)
{
  DISK_GEOMETRY dg;
  HANDLE h;
  int x,y,r=0;

  DisplayStatus(hWnd,"Reading disk");
  startsec=0;
  if (stopthread1)
    return FALSE;
	if (ask)
	{
		if (MessageBox(hWnd,"Insert SOURCE disk in drive A:","Insert Source disk",MB_OKCANCEL|MB_ICONEXCLAMATION)==IDCANCEL)
			return FALSE;
	}
  h=OpenInputFile(hWnd,"\\\\.\\A:");
  if (!h)
    return FALSE;
  DisplayStatus(hWnd,"Getting Disk Information");
  if (!GetDiskInfo(hWnd,h,&dg))
  {
    CloseHandle(h);
    return FALSE;
  }
  g_dg=dg;
  DisplayDiskInfo(hWnd,dg);
  DisplayStatus(hWnd,"Allocating Memory");
  if (!AllocateMemory(hWnd,dg))
  {
    CloseHandle(h);
    return FALSE;
  }
  DisplayStatus(hWnd,"Reading disk");
  startsec=time(NULL);
  for (x=0;x<dg.Cylinders.QuadPart && !stopthread1;x++)
  {
    for (y=0;y<(int)dg.TracksPerCylinder && !stopthread1;y++)
    {
      r=0;
      DisplayProgress(hWnd,x,y,r,dg);
      if (!fat_selection || x==0 || (x && fulltrack[x]))
      {
				while (r>=0 && !stopthread1 && !ReadTrack(h,x,y,&buffer[BUFPOS(x,y,dg)],dg))
				{
					if (++r>=retries)
					{
						switch (AskLastError(hWnd))
						{
							case IDABORT:
							CloseHandle(h);
							FreeMemory();
							return FALSE;
									case IDIGNORE:
							r=-1;
							break;
									case IDRETRY:
							CloseHandle(h);
							h=OpenInputFile(hWnd,"\\\\.\\A:");
							if (!h)
								return FALSE;
							r=0;
							break;
						}
					}
					DisplayProgress(hWnd,x,y,r,dg);
				}
				if (fat_selection && x==0 && y==1)
          GetEmptyTracks(buffer,dg);
      }
    }
  }
  CloseHandle(h);
  return TRUE;
}

int WriteDisk(HWND hWnd,int ask)
{
  HANDLE h;
  int x,y,r=0;
  DISK_GEOMETRY dg;
  void *verify_buffer=NULL;
  char str[70];

  DisplayStatus(hWnd,"Writing disk");
  startsec=0;
  if (stopthread1)
    return FALSE;
  if (!buffer)
  {
    MessageBox(hWnd,"No disk image loaded in memory","Error",MB_OK|MB_ICONEXCLAMATION);
    return FALSE;
  }
	if (ask)
	{
		if (MessageBox(hWnd,"Insert TARGET disk in drive A:","Insert Target disk",MB_OKCANCEL|MB_ICONEXCLAMATION)==IDCANCEL)
			return FALSE;
	}
  h=OpenOutputFile(hWnd,"\\\\.\\A:",OPEN_EXISTING);
  if (!h)
    return FALSE;
  DisplayDiskInfo(hWnd,g_dg);
  DisplayStatus(hWnd,"Getting Disk Information");
  if (!GetDiskInfo(hWnd,h,&dg))
  {
    CloseHandle(h);
    return FALSE;
  }
  if (!always_format && memcmp(&dg,&g_dg,sizeof(DISK_GEOMETRY))!=0)
  {
    w_dg=g_dg;
    if (!FormatDisk(hWnd,&h,FALSE))
    {
      CloseHandle(h);
      return FALSE;
    }
  }
  else if (always_format)
  {
    w_dg=g_dg;
    if (!FormatDisk(hWnd,&h,FALSE))
    {
      CloseHandle(h);
      return FALSE;
    }
  }
  if (verify_data)
  {
    if ((verify_buffer=GlobalAlloc(GMEM_FIXED,g_dg.SectorsPerTrack*g_dg.BytesPerSector))==NULL)
    {
      DisplayLastError(hWnd);
      CloseHandle(h);
      return FALSE;
    }
  }
  startsec=time(NULL);
  for (x=0;x<g_dg.Cylinders.QuadPart && !stopthread1;x++)
  {
    for (y=0;y<(int)g_dg.TracksPerCylinder && !stopthread1;y++)
    {
      r=0;
      DisplayProgress(hWnd,x,y,r,g_dg);
      if (!fat_selection || x==0 || (x && fulltrack[x]))
      {
	while (r>=0 && !stopthread1)
	{
          DisplayStatus(hWnd,"Writing disk");
	  if (!stopthread1 && !WriteTrack(h,x,y,&buffer[BUFPOS(x,y,g_dg)]))
	  {
	    if (r==0 && x==0 && y==0 && !always_format && GetLastError()!=19)
	    {
	      w_dg=g_dg;
	      if (!FormatDisk(hWnd,&h,FALSE))
	      {
		if (verify_data)
		  GlobalFree(verify_buffer);
		CloseHandle(h);
		return FALSE;
	      }
	      DisplayStatus(hWnd,"Writing disk");
	    }
	    if (++r>=retries)
	    {
	      switch (AskLastError(hWnd))
	      {
		case IDABORT:
		  if (verify_data)
		    GlobalFree(verify_buffer);
		  CloseHandle(h);
		  return FALSE;
		case IDIGNORE:
		  r=-1;
		  break;
		case IDRETRY:
		  CloseHandle(h);
		  h=OpenOutputFile(hWnd,"\\\\.\\A:",OPEN_EXISTING);
		  if (!h)
		  {
		    if (verify_data)
		      GlobalFree(verify_buffer);
		    return FALSE;
		  }
		  r=0;
		  break;
	      }
	    }
            DisplayProgress(hWnd,x,y,r,g_dg);
	    continue;
	  }
	  else
	  {
	    if (!verify_data)
	      r=-1;
	  }
	  if (verify_data)
	  {
	    DisplayStatus(hWnd,"Verifying disk");
	    if (!stopthread1 && !ReadTrack(h,x,y,verify_buffer,g_dg))
	    {
	      if (++r>=retries)
	      {
		switch (AskLastError(hWnd))
		{
		  case IDABORT:
		    if (verify_data)
		      GlobalFree(verify_buffer);
		    CloseHandle(h);
		    return FALSE;
		  case IDIGNORE:
		    r=-1;
		    break;
		  case IDRETRY:
		    r=0;
		    break;
		}
	      }
              DisplayProgress(hWnd,x,y,r,g_dg);
	      continue;
	    }
	    else
	      r=-1;
	  }
	}
	if (verify_data && !stopthread1)
	{
	  if (memcmp(verify_buffer,&buffer[BUFPOS(x,y,dg)],g_dg.SectorsPerTrack*g_dg.BytesPerSector))
	  {
	    sprintf(str,"Track %d, Head %d is different than the original",x,y);
	    if (MessageBox(hWnd,str,"Verify Error",MB_OKCANCEL|MB_ICONEXCLAMATION)==IDCANCEL)
	    {
	      GlobalFree(verify_buffer);
	      CloseHandle(h);
	      return FALSE;
	    }
	  }
	}
      }
    }
  }
  if (verify_data)
    GlobalFree(verify_buffer);
  CloseHandle(h);
  return TRUE;
}

int CompareDisk(HWND hWnd,int ask)
{
  HANDLE h;
  int x,y,r=0,different=0;
  DISK_GEOMETRY dg;
  void *verify_buffer=NULL;
  char str[70];

  DisplayStatus(hWnd,"Comparing disk");
  startsec=0;
  if (stopthread1)
    return FALSE;
  if (!buffer)
  {
    MessageBox(hWnd,"No disk image loaded in memory","Error",MB_OK|MB_ICONEXCLAMATION);
    return FALSE;
  }
	if (ask)
	{
		if (MessageBox(hWnd,"Insert SECOND disk in drive A:","Insert Second disk",MB_OKCANCEL|MB_ICONEXCLAMATION)==IDCANCEL)
			return FALSE;
	}
  h=OpenInputFile(hWnd,"\\\\.\\A:");
  if (!h)
    return FALSE;
  DisplayDiskInfo(hWnd,g_dg);
  DisplayStatus(hWnd,"Getting Disk Information");
  if (!GetDiskInfo(hWnd,h,&dg))
  {
    CloseHandle(h);
    return FALSE;
  }
  if (memcmp(&dg,&g_dg,sizeof(DISK_GEOMETRY))!=0)
  {
    MessageBox(hWnd,"Disks are not the same size","Error",MB_OKCANCEL);
    CloseHandle(h);
    return FALSE;
  }
  if ((verify_buffer=GlobalAlloc(GMEM_FIXED,g_dg.SectorsPerTrack*g_dg.BytesPerSector))==NULL)
  {
    DisplayLastError(hWnd);
    CloseHandle(h);
    return FALSE;
  }
  startsec=time(NULL);
  DisplayStatus(hWnd,"Comparing disk");
  for (x=0;x<g_dg.Cylinders.QuadPart && !stopthread1;x++)
  {
    for (y=0;y<(int)g_dg.TracksPerCylinder && !stopthread1;y++)
    {
      r=0;
      DisplayProgress(hWnd,x,y,r,g_dg);
      if (!fat_selection || x==0 || (x && fulltrack[x]))
      {
	while (r>=0 && !stopthread1 && !ReadTrack(h,x,y,verify_buffer,g_dg))
	{
	  if (++r>=retries)
	  {
	    switch (AskLastError(hWnd))
	    {
	      case IDABORT:
		if (verify_data)
		  GlobalFree(verify_buffer);
		CloseHandle(h);
		return FALSE;
	      case IDIGNORE:
		r=-1;
		break;
	      case IDRETRY:
		r=0;
		break;
	    }
	  }
          DisplayProgress(hWnd,x,y,r,g_dg);
	}
	if (!stopthread1)
	{
	  if (memcmp(verify_buffer,&buffer[BUFPOS(x,y,dg)],g_dg.SectorsPerTrack*g_dg.BytesPerSector))
	  {
	    sprintf(str,"Track %d, Head %d is different than the original",x,y);
	    if (MessageBox(hWnd,str,"Compare Error",MB_OKCANCEL|MB_ICONEXCLAMATION)==IDCANCEL)
	    {
	      GlobalFree(verify_buffer);
	      CloseHandle(h);
	      return FALSE;
	    }
	    different=1;
	  }
	}
      }
    }
  }
  if (verify_data)
    GlobalFree(verify_buffer);
  CloseHandle(h);
  if (!stopthread1)
  {
    if (!different)
      MessageBox(hWnd,"Disk is equal to the original","Compare complete",MB_OK|MB_ICONEXCLAMATION);
    else
      MessageBox(hWnd,"Disk is different than the original","Compare error",MB_OK|MB_ICONEXCLAMATION);
  }
  return TRUE;
}

int ReadImage(HWND hWnd)
{
  HANDLE h;
  DWORD bytes_read;
  char filename[512];
  DISK_GEOMETRY dg;

  if (stopthread1)
    return FALSE;
  GetDlgItemText(g_hWnd,IDC_SOURCE_FILE,filename,512);
  DisplayStatus(hWnd,"Opening file");
  h=OpenInputFile(hWnd,filename);
  if (!h)
    return FALSE;
  DisplayStatus(hWnd,"Getting file info");
  if (!GetDiskImageFormat(hWnd,h,&dg))
  {
    CloseHandle(h);
    return FALSE;
  }
  DisplayDiskInfo(hWnd,dg);
  DisplayStatus(hWnd,"Allocating memory");
  if (!AllocateMemory(hWnd,dg))
  {
    CloseHandle(h);
    return FALSE;
  }
  DisplayStatus(hWnd,"Reading disk image");
	if (!ReadFile(h,buffer,BUFSIZE(dg),&bytes_read,NULL))
	{
		DisplayLastError(hWnd);
		CloseHandle(h);
		FreeMemory();
		return FALSE;
	}
  if (fat_selection)
    GetEmptyTracks(buffer,dg);
  CloseHandle(h);
  return TRUE;
}

int CompareImage(HWND hWnd)
{
  HANDLE h;
  DWORD bytes_read;
  char filename[512];
  void *verify_buffer;

  if (stopthread1)
    return FALSE;
  if (!buffer)
  {
    MessageBox(hWnd,"No disk image loaded in memory","Error",MB_OK|MB_ICONEXCLAMATION);
    return FALSE;
  }
  GetDlgItemText(g_hWnd,IDC_TARGET_FILE,filename,512);
  DisplayStatus(hWnd,"Opening file");
  h=OpenInputFile(hWnd,filename);
  if (!h)
    return FALSE;
  DisplayStatus(hWnd,"Allocating memory");
  if ((verify_buffer=GlobalAlloc(GMEM_FIXED,BUFSIZE(g_dg)))==NULL)
  {
    DisplayLastError(hWnd);
    CloseHandle(h);
    return FALSE;
  }
  DisplayStatus(hWnd,"Comparing disk image");
  if (!ReadFile(h,verify_buffer,BUFSIZE(g_dg),&bytes_read,NULL))
  {
    DisplayLastError(hWnd);
    CloseHandle(h);
    GlobalFree(verify_buffer);
    return FALSE;
  }
  if (memcmp(buffer,verify_buffer,BUFSIZE(g_dg)))
    MessageBox(hWnd,"Disk image is different than the original","Compare error",MB_OK|MB_ICONEXCLAMATION);
  else
    MessageBox(hWnd,"Disk image is equal to the original","Compare complete",MB_OK|MB_ICONEXCLAMATION);
  CloseHandle(h);
  GlobalFree(verify_buffer);
  return TRUE;
}

int WriteImage(HWND hWnd)
{
  HANDLE h;
  DWORD bytes_written;
  char filename[512];

  if (stopthread1)
    return FALSE;
  if (!buffer)
  {
    MessageBox(hWnd,"No disk image loaded in memory","Error",MB_OK|MB_ICONEXCLAMATION);
    return FALSE;
  }
  GetDlgItemText(g_hWnd,IDC_TARGET_FILE,filename,512);
  DisplayStatus(hWnd,"Creating file");
  h=OpenOutputFile(hWnd,filename,CREATE_ALWAYS);
  if (!h)
    return FALSE;
  DisplayStatus(hWnd,"Writing disk image");
	if (!WriteFile(h,buffer,BUFSIZE(g_dg),&bytes_written,NULL))
	{
		DisplayLastError(hWnd);
		CloseHandle(h);
		return FALSE;
	}
  CloseHandle(h);
  return TRUE;
}

DWORD WINAPI CopyDiskToDisk(LPVOID hWnd)
{
  DisplayStatus(hWnd,"Initializing");
  if (ReadDisk(hWnd,0))
    WriteDisk(hWnd,1);
  SendMessage(hWnd,WM_USER,THREAD_END,0);
  ExitThread(0);
  return TRUE;
}

DWORD WINAPI CopyDiskToImage(LPVOID hWnd)
{
  DisplayStatus(hWnd,"Initializing");
  if (ReadDisk(hWnd,0))
    WriteImage(hWnd);
  SendMessage(hWnd,WM_USER,THREAD_END,0);
  ExitThread(0);
  return TRUE;
}

DWORD WINAPI CopyImageToDisk(LPVOID hWnd)
{
  DisplayStatus(hWnd,"Initializing");
  if (ReadImage(hWnd))
    WriteDisk(hWnd,0);
  SendMessage(hWnd,WM_USER,THREAD_END,0);
  ExitThread(0);
  return TRUE;
}

DWORD WINAPI CopyImageToImage(LPVOID hWnd)
{
  DisplayStatus(hWnd,"Initializing");
  if (ReadImage(hWnd))
    WriteImage(hWnd);
  SendMessage(hWnd,WM_USER,THREAD_END,0);
  ExitThread(0);
  return TRUE;
}

DWORD WINAPI CopyDiskToMemory(LPVOID hWnd)
{
  DisplayStatus(hWnd,"Initializing");
  ReadDisk(hWnd,0);
  SendMessage(hWnd,WM_USER,THREAD_END,0);
  ExitThread(0);
  return TRUE;
}

DWORD WINAPI CopyImageToMemory(LPVOID hWnd)
{
  DisplayStatus(hWnd,"Initializing");
  ReadImage(hWnd);
  SendMessage(hWnd,WM_USER,THREAD_END,0);
  ExitThread(0);
  return TRUE;
}

DWORD WINAPI CopyMemoryToDisk(LPVOID hWnd)
{
  DisplayStatus(hWnd,"Initializing");
  WriteDisk(hWnd,0);
  SendMessage(hWnd,WM_USER,THREAD_END,0);
  ExitThread(0);
  return TRUE;
}

DWORD WINAPI CopyMemoryToImage(LPVOID hWnd)
{
  DisplayStatus(hWnd,"Initializing");
  WriteImage(hWnd);
  SendMessage(hWnd,WM_USER,THREAD_END,0);
  ExitThread(0);
  return TRUE;
}

DWORD WINAPI FormatTargetDisk(LPVOID hWnd)
{
  HANDLE h;

  DisplayStatus(hWnd,"Initializing");
  if (AskDiskFormat(hWnd,1))
  {
    DisplayDiskInfo(hWnd,w_dg);
/*    if (MessageBox(hWnd,"Insert TARGET disk in drive A:","Insert Target disk",MB_OKCANCEL|MB_ICONEXCLAMATION)==IDCANCEL)
    {
      SendMessage(hWnd,WM_USER,THREAD_END,0);
      ExitThread(0);
      return FALSE;
    }*/
    h=OpenOutputFile(hWnd,"\\\\.\\A:",OPEN_EXISTING);
    if (!h)
    {
      SendMessage(hWnd,WM_USER,THREAD_END,0);
      ExitThread(0);
      return FALSE;
    }
    FormatDisk(hWnd,&h,TRUE);
    CloseHandle(h);
  }
  SendMessage(hWnd,WM_USER,THREAD_END,0);
  ExitThread(0);
  return TRUE;
}

DWORD WINAPI CompareDiskToDisk(LPVOID hWnd)
{
  DisplayStatus(hWnd,"Initializing");
  if (ReadDisk(hWnd,0))
    CompareDisk(hWnd,1);
  SendMessage(hWnd,WM_USER,THREAD_END,0);
  ExitThread(0);
  return TRUE;
}

DWORD WINAPI CompareDiskToImage(LPVOID hWnd)
{
  DisplayStatus(hWnd,"Initializing");
  if (ReadDisk(hWnd,0))
    CompareImage(hWnd);
  SendMessage(hWnd,WM_USER,THREAD_END,0);
  ExitThread(0);
  return TRUE;
}

DWORD WINAPI CompareImageToDisk(LPVOID hWnd)
{
  DisplayStatus(hWnd,"Initializing");
  if (ReadImage(hWnd))
    CompareDisk(hWnd,0);
  SendMessage(hWnd,WM_USER,THREAD_END,0);
  ExitThread(0);
  return TRUE;
}

DWORD WINAPI CompareImageToImage(LPVOID hWnd)
{
  DisplayStatus(hWnd,"Initializing");
  if (ReadImage(hWnd))
    CompareImage(hWnd);
  SendMessage(hWnd,WM_USER,THREAD_END,0);
  ExitThread(0);
  return TRUE;
}

DWORD WINAPI CompareMemoryToDisk(LPVOID hWnd)
{
  DisplayStatus(hWnd,"Initializing");
  CompareDisk(hWnd,0);
  SendMessage(hWnd,WM_USER,THREAD_END,0);
  ExitThread(0);
  return TRUE;
}

DWORD WINAPI CompareMemoryToImage(LPVOID hWnd)
{
  DisplayStatus(hWnd,"Initializing");
  CompareImage(hWnd);
  SendMessage(hWnd,WM_USER,THREAD_END,0);
  ExitThread(0);
  return TRUE;
}


